// Andrew Naplavkov

#define CATCH_CONFIG_COUNTER
#define CATCH_CONFIG_ENABLE_BENCHMARKING
#define CATCH_CONFIG_MAIN
#include <catch.hpp>

#include <step/test/edit_distance.hpp>
#include <step/test/kahan.hpp>
#include <step/test/longest_common_subsequence.hpp>
#include <step/test/longest_common_substring.hpp>
#include <step/test/longest_increasing_subsequence.hpp>
#include <step/test/longest_repeated_substring.hpp>
#include <step/test/maximum_subarray.hpp>
#include <step/test/suffix.hpp>
